from __future__ import annotations

import itertools
from collections.abc import Callable, Generator, Iterable, Iterator
from typing import TYPE_CHECKING, Any, overload

import cytoolz as cz
import more_itertools as mit

from .._core import IterWrapper

if TYPE_CHECKING:
    from ._main import Iter


class BaseJoins[T](IterWrapper[T]):
    @overload
    def zip[T1](
        self, iter1: Iterable[T1], /, *, strict: bool = ...
    ) -> Iter[tuple[T, T1]]: ...
    @overload
    def zip[T1, T2](
        self,
        iter1: Iterable[T1],
        iter2: Iterable[T2],
        /,
        *,
        strict: bool = ...,
    ) -> Iter[tuple[T, T1, T2]]: ...
    @overload
    def zip[T1, T2, T3](
        self,
        iter1: Iterable[T1],
        iter2: Iterable[T2],
        iter3: Iterable[T3],
        /,
        *,
        strict: bool = ...,
    ) -> Iter[tuple[T, T1, T2, T3]]: ...
    @overload
    def zip[T1, T2, T3, T4](
        self,
        iter1: Iterable[T1],
        iter2: Iterable[T2],
        iter3: Iterable[T3],
        iter4: Iterable[T4],
        /,
        *,
        strict: bool = ...,
    ) -> Iter[tuple[T, T1, T2, T3, T4]]: ...
    def zip(
        self, *others: Iterable[Any], strict: bool = False
    ) -> Iter[tuple[Any, ...]]:
        """
        Zip with other iterables, optionally strict.

        Args:
            *others: Other iterables to zip with.
            strict: Whether to enforce equal lengths of iterables. Defaults to False.
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2]).zip([10, 20]).into(list)
        [(1, 10), (2, 20)]
        >>> pc.Iter.from_(["a", "b"]).zip([1, 2, 3]).into(list)
        [('a', 1), ('b', 2)]

        ```
        """
        return self.apply(zip, *others, strict=strict)

    def zip_offset[U](
        self,
        *others: Iterable[T],
        offsets: list[int],
        longest: bool = False,
        fillvalue: U = None,
    ) -> Iter[tuple[T | U, ...]]:
        """
        Zip the input iterables together, but offset the i-th iterable by the i-th item in offsets.

        Args:
            *others: Other iterables to zip with.
            offsets: List of integers specifying the offsets for each iterable.
            longest: Whether to continue until the longest iterable is exhausted. Defaults to False.
            fillvalue: Value to use for missing elements. Defaults to None.
        Example:
        ```python
        >>> import pyochain as pc
        >>> data = pc.Seq("0123")
        >>> data.iter().zip_offset("abcdef", offsets=(0, 1)).into(list)
        [('0', 'b'), ('1', 'c'), ('2', 'd'), ('3', 'e')]

        ```
        This can be used as a lightweight alternative to SciPy or pandas to analyze data sets in which some series have a lead or lag relationship.

        By default, the sequence will end when the shortest iterable is exhausted.

        To continue until the longest iterable is exhausted, set longest to True.
        ```python
        >>> data.iter().zip_offset("abcdef", offsets=(0, 1), longest=True).into(list)
        [('0', 'b'), ('1', 'c'), ('2', 'd'), ('3', 'e'), (None, 'f')]

        ```
        """

        def _zip_offset(data: Iterable[T]) -> Iterator[tuple[T | U, ...]]:
            return mit.zip_offset(
                data,
                *others,
                offsets=offsets,
                longest=longest,
                fillvalue=fillvalue,
            )

        return self.apply(_zip_offset)

    @overload
    def zip_broadcast[T1](
        self,
        iter1: Iterable[T1],
        /,
        *,
        strict: bool = False,
    ) -> Iter[tuple[T, T1]]: ...
    @overload
    def zip_broadcast[T1, T2](
        self,
        iter1: Iterable[T1],
        iter2: Iterable[T2],
        /,
        *,
        strict: bool = False,
    ) -> Iter[tuple[T, T1, T2]]: ...
    @overload
    def zip_broadcast[T1, T2, T3](
        self,
        iter1: Iterable[T1],
        iter2: Iterable[T2],
        iter3: Iterable[T3],
        /,
        *,
        strict: bool = False,
    ) -> Iter[tuple[T, T1, T2, T3]]: ...
    @overload
    def zip_broadcast[T1, T2, T3, T4](
        self,
        iter1: Iterable[T1],
        iter2: Iterable[T2],
        iter3: Iterable[T3],
        iter4: Iterable[T4],
        /,
        *,
        strict: bool = False,
    ) -> Iter[tuple[T, T1, T2, T3, T4]]: ...
    def zip_broadcast(
        self, *others: Iterable[Any], strict: bool = False
    ) -> Iter[tuple[Any, ...]]:
        """
        Version of zip that "broadcasts" any scalar (i.e., non-iterable) items into output tuples.

        `str` and `bytes` are not treated as iterables.

        If the strict keyword argument is True, then UnequalIterablesError will be raised if any of the iterables have different lengths.
        Args:
            *others: Other iterables or scalars to zip with.
            strict: Whether to enforce equal lengths of iterables. Defaults to False.
        Example:
        ```python
        >>> import pyochain as pc
        >>> data = pc.Iter.from_([1, 2, 3])
        >>> other = ["a", "b", "c"]
        >>> scalar = "_"
        >>> data.zip_broadcast(other, scalar).into(list)
        [(1, 'a', '_'), (2, 'b', '_'), (3, 'c', '_')]

        ```
        """

        def _zip_broadcast(
            *objects: Iterable[Any],
        ) -> Generator[tuple[Iterable[Any], ...] | tuple[object, ...], Any, None]:
            """from more_itertools.zip_broadcast"""

            def is_scalar(obj: Any) -> bool:
                if isinstance(obj, (str, bytes)):
                    return True
                try:
                    iter(obj)
                except TypeError:
                    return True
                else:
                    return False

            size = len(objects)
            if not size:
                return

            new_item: list[object] = [None] * size
            iterables: list[Iterator[Any]] = []
            iterable_positions: list[int] = []
            for i, obj in enumerate(objects):
                if is_scalar(obj):
                    new_item[i] = obj
                else:
                    iterables.append(iter(obj))
                    iterable_positions.append(i)

            if not iterables:
                yield tuple(objects)
                return

            zipper = mit.zip_equal if strict else zip
            for item in zipper(*iterables):
                for i, new_item[i] in zip(iterable_positions, item):
                    pass
                yield tuple(new_item)

        return self.apply(_zip_broadcast, *others)

    @overload
    def zip_equal(self) -> Iter[tuple[T]]: ...
    @overload
    def zip_equal[T2](self, __iter2: Iterable[T2]) -> Iter[tuple[T, T2]]: ...
    @overload
    def zip_equal[T2, T3](
        self, __iter2: Iterable[T2], __iter3: Iterable[T3]
    ) -> Iter[tuple[T, T2, T3]]: ...
    @overload
    def zip_equal[T2, T3, T4](
        self,
        __iter2: Iterable[T2],
        __iter3: Iterable[T3],
        __iter4: Iterable[T4],
    ) -> Iter[tuple[T, T2, T3, T4]]: ...
    @overload
    def zip_equal[T2, T3, T4, T5](
        self,
        __iter2: Iterable[T2],
        __iter3: Iterable[T3],
        __iter4: Iterable[T4],
        __iter5: Iterable[T5],
    ) -> Iter[tuple[T, T2, T3, T4, T5]]: ...
    def zip_equal(self, *others: Iterable[Any]) -> Iter[tuple[Any, ...]]:
        """
        `zip` the input *iterables* together but raise `UnequalIterablesError` if they aren't all the same length.

        Args:
            *others: Other iterables to zip with.
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_(range(3)).zip_equal("abc").into(list)
        [(0, 'a'), (1, 'b'), (2, 'c')]
        >>> pc.Iter.from_(range(3)).zip_equal("abcd").into(list)
        ... # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        ...
        more_itertools.more.UnequalIterablesError: Iterables have different
        lengths

        ```
        """

        def _zip_equal(data: Iterable[T]) -> Iterator[tuple[Any, ...]]:
            return mit.zip_equal(data, *others)

        return self.apply(_zip_equal)

    def zip_longest[U](
        self, *others: Iterable[T], fill_value: U = None
    ) -> Iter[tuple[U | T, ...]]:
        """
        Zip with other iterables, filling missing values.

        Args:
            *others: Other iterables to zip with.
            fill_value: Value to use for missing elements. Defaults to None.
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2]).zip_longest([10], fill_value=0).into(list)
        [(1, 10), (2, 0)]

        ```
        """
        return self.apply(itertools.zip_longest, *others, fillvalue=fill_value)

    @overload
    def product(self) -> Iter[tuple[T]]: ...
    @overload
    def product[T1](self, iter1: Iterable[T1], /) -> Iter[tuple[T, T1]]: ...
    @overload
    def product[T1, T2](
        self, iter1: Iterable[T1], iter2: Iterable[T2], /
    ) -> Iter[tuple[T, T1, T2]]: ...
    @overload
    def product[T1, T2, T3](
        self, iter1: Iterable[T1], iter2: Iterable[T2], iter3: Iterable[T3], /
    ) -> Iter[tuple[T, T1, T2, T3]]: ...
    @overload
    def product[T1, T2, T3, T4](
        self,
        iter1: Iterable[T1],
        iter2: Iterable[T2],
        iter3: Iterable[T3],
        iter4: Iterable[T4],
        /,
    ) -> Iter[tuple[T, T1, T2, T3, T4]]: ...

    def product(self, *others: Iterable[Any]) -> Iter[tuple[Any, ...]]:
        """
        Computes the Cartesian product with another iterable.
        This is the declarative equivalent of nested for-loops.

        It pairs every element from the source iterable with every element from the
        other iterable.

        Args:
            *others: Other iterables to compute the Cartesian product with.
        Example:
        ```python
        >>> import pyochain as pc
        >>> colors = pc.Iter.from_(["blue", "red"])
        >>> sizes = ["S", "M"]
        >>> colors.product(sizes).into(list)
        [('blue', 'S'), ('blue', 'M'), ('red', 'S'), ('red', 'M')]

        ```
        """
        return self.apply(itertools.product, *others)

    def diff_at(
        self,
        *others: Iterable[T],
        default: T | None = None,
        key: Callable[[T], Any] | None = None,
    ) -> Iter[tuple[T, ...]]:
        """
        Return those items that differ between iterables.
        Each output item is a tuple where the i-th element is from the i-th input iterable.

        If an input iterable is exhausted before others, then the corresponding output items will be filled with *default*.
        Args:
            *others: Other iterables to compare with.
            default: Value to use for missing elements. Defaults to None.
            key: Function to apply to each item for comparison. Defaults to None.
        Example:
        ```python
        >>> import pyochain as pc
        >>> data = pc.Seq([1, 2, 3])
        >>> data.iter().diff_at([1, 2, 10, 100], default=None).into(list)
        [(3, 10), (None, 100)]
        >>> data.iter().diff_at([1, 2, 10, 100, 2, 6, 7], default=0).into(list)
        [(3, 10), (0, 100), (0, 2), (0, 6), (0, 7)]

        A key function may also be applied to each item to use during comparisons:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_(["apples", "bananas"]).diff_at(
        ...     ["Apples", "Oranges"], key=str.lower
        ... ).into(list)
        [('bananas', 'Oranges')]

        ```
        """
        return self.apply(cz.itertoolz.diff, *others, default=default, key=key)

    def join[R, K](
        self,
        other: Iterable[R],
        left_on: Callable[[T], K],
        right_on: Callable[[R], K],
        left_default: T | None = None,
        right_default: R | None = None,
    ) -> Iter[tuple[T, R]]:
        """
        Perform a relational join with another iterable.

        Args:
            other: Iterable to join with.
            left_on: Function to extract the join key from the left iterable.
            right_on: Function to extract the join key from the right iterable.
            left_default: Default value for missing elements in the left iterable. Defaults to None.
            right_default: Default value for missing elements in the right iterable. Defaults to None.
        Example:
        ```python
        >>> import pyochain as pc
        >>> colors = pc.Iter.from_(["blue", "red"])
        >>> sizes = ["S", "M"]
        >>> colors.join(sizes, left_on=lambda c: c, right_on=lambda s: s).into(list)
        [(None, 'S'), (None, 'M'), ('blue', None), ('red', None)]

        ```
        """

        def _join(data: Iterable[T]) -> Iterator[tuple[T, R]]:
            return cz.itertoolz.join(
                leftkey=left_on,
                leftseq=data,
                rightkey=right_on,
                rightseq=other,
                left_default=left_default,
                right_default=right_default,
            )

        return self.apply(_join)



from ._main import CommonBase, IterWrapper, MappingWrapper, Pipeable, Wrapper
from ._protocols import (
    Peeked,
    SizedIterable,
    SupportsAllComparisons,
    SupportsKeysAndGetItem,
    SupportsRichComparison,
)

__all__ = [
    "MappingWrapper",
    "CommonBase",
    "IterWrapper",
    "Wrapper",
    "SupportsAllComparisons",
    "SupportsRichComparison",
    "SupportsKeysAndGetItem",
    "Peeked",
    "SizedIterable",
    "Pipeable",
]



from __future__ import annotations

from collections.abc import Callable, Iterable
from typing import TYPE_CHECKING

import rolling

from .._core import IterWrapper

if TYPE_CHECKING:
    from ._main import Iter


class BaseRolling[T](IterWrapper[T]):
    def rolling_mean(self, window_size: int) -> Iter[float]:
        """
        Compute the rolling mean.

        Args:
            window_size: Size of the rolling window.

        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 3, 4, 5]).rolling_mean(3).into(list)
        [2.0, 3.0, 4.0]

        ```
        """
        return self.apply(rolling.Mean, window_size)

    def rolling_median(self, window_size: int) -> Iter[T]:
        """
        Compute the rolling median.

        Args:
            window_size: Size of the rolling window.

        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 3, 2, 5, 4]).rolling_median(3).into(list)
        [2, 3, 4]

        ```
        """
        return self.apply(rolling.Median, window_size)

    def rolling_sum(self, window_size: int) -> Iter[T]:
        """
        Compute the rolling sum.

        Args:
            window_size: Size of the rolling window.

        Will return integers if the input is integers, floats if the input is floats or mixed.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1.0, 2, 3, 4, 5]).rolling_sum(3).into(list)
        [6.0, 9.0, 12.0]

        ```
        """
        return self.apply(rolling.Sum, window_size)

    def rolling_min(self, window_size: int) -> Iter[T]:
        """
        Compute the rolling minimum.

        Args:
            window_size: Size of the rolling window.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([3, 1, 4, 1, 5, 9, 2]).rolling_min(3).into(list)
        [1, 1, 1, 1, 2]

        ```
        """
        return self.apply(rolling.Min, window_size)

    def rolling_max(self, window_size: int) -> Iter[T]:
        """
        Compute the rolling maximum.

        Args:
            window_size: Size of the rolling window.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([3, 1, 4, 1, 5, 9, 2]).rolling_max(3).into(list)
        [4, 4, 5, 9, 9]

        ```
        """
        return self.apply(rolling.Max, window_size)

    def rolling_var(self, window_size: int) -> Iter[float]:
        """
        Compute the rolling variance.

        Args:
            window_size: Size of the rolling window.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 4, 1, 4]).rolling_var(3).round(2).into(list)
        [2.33, 2.33, 3.0]

        ```
        """
        return self.apply(rolling.Var, window_size)

    def rolling_std(self, window_size: int) -> Iter[float]:
        """
        Compute the rolling standard deviation.

        Args:
            window_size: Size of the rolling window.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 4, 1, 4]).rolling_std(3).round(2).into(list)
        [1.53, 1.53, 1.73]

        ```
        """
        return self.apply(rolling.Std, window_size)

    def rolling_kurtosis(self, window_size: int) -> Iter[float]:
        """
        Compute the rolling kurtosis.

        Args:
            window_size: Size of the rolling window.

        The windows must have at least 4 observations.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 4, 1, 4]).rolling_kurtosis(4).into(list)
        [1.5, -3.901234567901234]

        ```
        """
        return self.apply(rolling.Kurtosis, window_size)

    def rolling_skew(self, window_size: int) -> Iter[float]:
        """
        Compute the rolling skewness.

        Args:
            window_size: Size of the rolling window.

        The windows must have at least 3 observations.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 4, 1, 4]).rolling_skew(3).round(2).into(list)
        [0.94, 0.94, -1.73]

        ```
        """
        return self.apply(rolling.Skew, window_size)

    def rolling_all(self, window_size: int) -> Iter[bool]:
        """
        Compute whether all values in the window evaluate to True.

        Args:
            window_size: Size of the rolling window.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([True, True, False, True, True]).rolling_all(2).into(list)
        [True, False, False, True]

        ```
        """
        return self.apply(rolling.All, window_size)

    def rolling_any(self, window_size: int) -> Iter[bool]:
        """
        Compute whether any value in the window evaluates to True.

        Args:
            window_size: Size of the rolling window.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([True, True, False, True, True]).rolling_any(2).into(list)
        [True, True, True, True]

        ```
        """
        return self.apply(rolling.Any, window_size)

    def rolling_product(self, window_size: int) -> Iter[float]:
        """
        Compute the rolling product.

        Args:
            window_size: Size of the rolling window.
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 3, 4, 5]).rolling_product(3).into(list)
        [6.0, 24.0, 60.0]

        ```
        """
        return self.apply(rolling.Product, window_size)

    def rolling_apply[R](
        self, func: Callable[[Iterable[T]], R], window_size: int
    ) -> Iter[R]:
        """
        Apply a custom function to each rolling window.

        Args:
            func: Function to apply to each rolling window.
            window_size: Size of the rolling window.

        The function should accept an iterable and return a single value.
        ```python
        >>> import pyochain as pc
        >>> def range_func(window):
        ...     return max(window) - min(window)
        >>> pc.Iter.from_([1, 3, 2, 5, 4]).rolling_apply(range_func, 3).into(list)
        [2, 3, 3]

        ```
        """
        return self.apply(rolling.Apply, window_size, "fixed", func)

    def rolling_apply_pairwise[R](
        self, other: Iterable[T], func: Callable[[T, T], R], window_size: int
    ) -> Iter[R]:
        """
        Apply a custom pairwise function to each rolling window of size 2.

        Args:
            other: Second iterable to apply the pairwise function.
            func: Function to apply to each pair of elements.
            window_size: Size of the rolling window.

        The function should accept two arguments and return a single value.
        ```python
        >>> import pyochain as pc
        >>> from statistics import correlation as corr
        >>> seq_1 = [1, 2, 3, 4, 5]
        >>> seq_2 = [1, 2, 3, 2, 1]
        >>> pc.Iter.from_(seq_1).rolling_apply_pairwise(seq_2, corr, 3).into(list)
        [1.0, 0.0, -1.0]

        ```
        """
        return self.apply(rolling.ApplyPairwise, other, window_size, func)



from __future__ import annotations

import itertools
from collections.abc import Callable, Iterable, Iterator
from typing import TYPE_CHECKING

import cytoolz as cz

if TYPE_CHECKING:
    from ._main import Iter


class IterConstructors:
    @staticmethod
    def from_count(start: int = 0, step: int = 1) -> Iter[int]:
        """
        Create an infinite iterator of evenly spaced values.

        **Warning** ⚠️
            This creates an infinite iterator.
            Be sure to use `Iter.take()` or `Iter.slice()` to limit the number of items taken.

        Args:
            start: Starting value of the sequence. Defaults to 0.
            step: Difference between consecutive values. Defaults to 1.
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_count(10, 2).take(3).into(list)
        [10, 12, 14]

        ```
        """
        from ._main import Iter

        return Iter(itertools.count(start, step))

    @staticmethod
    def from_func[U](func: Callable[[U], U], input: U) -> Iter[U]:
        """
        Create an infinite iterator by repeatedly applying a function on an original input.

        **Warning** ⚠️
            This creates an infinite iterator.
            Be sure to use `Iter.take()` or `Iter.slice()` to limit the number of items taken.

        Args:
            func: Function to apply repeatedly.
            input: Initial value to start the iteration.

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_func(lambda x: x + 1, 0).take(3).into(list)
        [0, 1, 2]

        ```
        """
        from ._main import Iter

        return Iter(cz.itertoolz.iterate(func, input))

    @staticmethod
    def from_[U](data: Iterable[U]) -> Iter[U]:
        """
        Create an iterator from any Iterable.

        - An Iterable is any object capable of returning its members one at a time, permitting it to be iterated over in a for-loop.
        - An Iterator is an object representing a stream of data; returned by calling `iter()` on an Iterable.
        - Once an Iterator is exhausted, it cannot be reused or reset.

        If you need to reuse the data, consider collecting it into a list first with `.collect()`.

        In general, avoid intermediate references when dealing with lazy iterators, and prioritize method chaining instead.
        Args:
            data: Iterable to convert into an iterator.
        Example:
        ```python
        >>> import pyochain as pc
        >>> data: tuple[int, ...] = (1, 2, 3)
        >>> iterator = pc.Iter.from_(data)
        >>> iterator.unwrap().__class__.__name__
        'tuple_iterator'
        >>> mapped = iterator.map(lambda x: x * 2)
        >>> mapped.unwrap().__class__.__name__
        'map'
        >>> mapped.collect(tuple).unwrap()
        (2, 4, 6)
        >>> # iterator is now exhausted
        >>> iterator.collect().unwrap()
        []

        ```
        """
        from ._main import Iter

        return Iter(iter(data))

    @staticmethod
    def unfold[S, V](seed: S, generator: Callable[[S], tuple[V, S] | None]) -> Iter[V]:
        """
        Create an iterator by repeatedly applying a generator function to an initial state.

        The `generator` function takes the current state and must return:
            - A tuple `(value, new_state)` to emit the `value` and continue with the `new_state`.
            - `None` to stop the generation.

        This is functionally equivalent to a state-based `while` loop.

        **Warning** ⚠️
            If the `generator` function never returns `None`, it creates an infinite iterator.
            Be sure to use `Iter.take()` or `Iter.slice()` to limit the number of items taken if necessary.

        Args:
            seed: Initial state for the generator.
            generator: Function that generates the next value and state.

        Example:
        ```python
        >>> import pyochain as pc
        >>> # Example 1: Simple counter up to 5
        >>> def counter_generator(state: int) -> tuple[int, int] | None:
        ...     if state < 5:
        ...         return (state * 10, state + 1)
        ...     return None
        >>> pc.Iter.unfold(seed=0, generator=counter_generator).into(list)
        [0, 10, 20, 30, 40]
        >>> # Example 2: Fibonacci sequence up to 100
        >>> type FibState = tuple[int, int]
        >>> def fib_generator(state: FibState) -> tuple[int, FibState] | None:
        ...     a, b = state
        ...     if a > 100:
        ...         return None
        ...     return (a, (b, a + b))
        >>> pc.Iter.unfold(seed=(0, 1), generator=fib_generator).into(list)
        [0, 1, 1, 2, 3, 5, 8, 13, 21, 34, 55, 89]
        >>> # Example 3: Infinite iterator (requires take())
        >>> pc.Iter.unfold(seed=1, generator=lambda s: (s, s * 2)).take(5).into(list)
        [1, 2, 4, 8, 16]

        ```
        """
        from ._main import Iter

        def _unfold() -> Iterator[V]:
            current_seed: S = seed
            while True:
                result: tuple[V, S] | None = generator(current_seed)
                if result is None:
                    break
                value, next_seed = result
                yield value
                current_seed = next_seed

        return Iter(_unfold())



from __future__ import annotations

import itertools
from collections.abc import Callable, Generator, Iterable, Iterator
from functools import partial
from typing import TYPE_CHECKING, Any, TypeGuard

import cytoolz as cz
import more_itertools as mit

from .._core import IterWrapper

if TYPE_CHECKING:
    from ._main import Iter


class BaseFilter[T](IterWrapper[T]):
    def filter(self, func: Callable[[T], bool]) -> Iter[T]:
        """
        Return an iterator yielding those items of iterable for which function is true.

        Args:
            func: Function to evaluate each item.
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 3]).filter(lambda x: x > 1).into(list)
        [2, 3]

        ```
        """

        def _filter(data: Iterable[T]) -> Iterator[T]:
            return (x for x in data if func(x))

        return self.apply(_filter)

    def filter_isin(self, values: Iterable[T]) -> Iter[T]:
        """
        Return elements that are in the given values iterable.

        Args:
            values: Iterable of values to check membership against.
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 3, 4]).filter_isin([2, 4, 6]).into(list)
        [2, 4]

        ```
        """

        def _filter_isin(data: Iterable[T]) -> Generator[T, None, None]:
            value_set: set[T] = set(values)
            return (x for x in data if x in value_set)

        return self.apply(_filter_isin)

    def filter_notin(self, values: Iterable[T]) -> Iter[T]:
        """
        Return elements that are not in the given values iterable.

        Args:
            values: Iterable of values to exclude.
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 3, 4]).filter_notin([2, 4, 6]).into(list)
        [1, 3]

        ```
        """

        def _filter_notin(data: Iterable[T]) -> Generator[T, None, None]:
            value_set: set[T] = set(values)
            return (x for x in data if x not in value_set)

        return self.apply(_filter_notin)

    def filter_contain(
        self: IterWrapper[str], text: str, format: Callable[[str], str] | None = None
    ) -> Iter[str]:
        """
        Return elements that contain the given text.

        Optionally, a format function can be provided to preprocess each element before checking for the substring.
        Args:
            text: Substring to check for.
            format: Optional function to preprocess each element before checking. Defaults to None.
        Example:
        ```python
        >>> import pyochain as pc
        >>>
        >>> data = pc.Seq(["apple", "banana", "cherry", "date"])
        >>> data.iter().filter_contain("ana").into(list)
        ['banana']
        >>> data.iter().map(str.upper).filter_contain("ana", str.lower).into(list)
        ['BANANA']

        ```
        """

        def _filter_contain(data: Iterable[str]) -> Generator[str, None, None]:
            def _(x: str) -> bool:
                formatted = format(x) if format else x
                return text in formatted

            return (x for x in data if _(x))

        return self.apply(_filter_contain)

    def filter_attr[U](self, attr: str, dtype: type[U] = object) -> Iter[U]:
        """
        Return elements that have the given attribute.

        The provided dtype is not checked at runtime for performance considerations.

        Args:
            attr: Name of the attribute to check for.
            dtype: Expected type of the attribute. Defaults to object.
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_(["hello", "world", 2, 5]).filter_attr("capitalize", str).into(
        ...     list
        ... )
        ['hello', 'world']

        ```
        """

        def check(data: Iterable[Any]) -> Generator[U, None, None]:
            def _(x: Any) -> TypeGuard[U]:
                return hasattr(x, attr)

            return (x for x in data if _(x))

        return self.apply(check)

    def filter_false(self, func: Callable[[T], bool]) -> Iter[T]:
        """
        Return elements for which func is false.

        Args:
            func: Function to evaluate each item.
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 3]).filter_false(lambda x: x > 1).into(list)
        [1]

        ```
        """
        return self.apply(partial(itertools.filterfalse, func))

    def filter_except(
        self, func: Callable[[T], object], *exceptions: type[BaseException]
    ) -> Iter[T]:
        """
        Yield the items from iterable for which the validator function does not raise one of the specified exceptions.

        Validator is called for each item in iterable.

        It should be a function that accepts one argument and raises an exception if that item is not valid.

        If an exception other than one given by exceptions is raised by validator, it is raised like normal.

        Args:
            func: Validator function to apply to each item.
            exceptions: Exceptions to catch and ignore.
        Example:
        ```python
        >>> import pyochain as pc
        >>> iterable = ["1", "2", "three", "4", None]
        >>> pc.Iter.from_(iterable).filter_except(int, ValueError, TypeError).into(list)
        ['1', '2', '4']

        ```
        """

        def _filter_except(data: Iterable[T]) -> Iterator[T]:
            return mit.filter_except(func, data, *exceptions)

        return self.apply(_filter_except)

    def take_while(self, predicate: Callable[[T], bool]) -> Iter[T]:
        """
        Take items while predicate holds.

        Args:
            predicate: Function to evaluate each item.
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 0]).take_while(lambda x: x > 0).into(list)
        [1, 2]

        ```
        """
        return self.apply(partial(itertools.takewhile, predicate))

    def skip_while(self, predicate: Callable[[T], bool]) -> Iter[T]:
        """
        Drop items while predicate holds.

        Args:
            predicate: Function to evaluate each item.
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 0]).skip_while(lambda x: x > 0).into(list)
        [0]

        ```
        """
        return self.apply(partial(itertools.dropwhile, predicate))

    def compress(self, *selectors: bool) -> Iter[T]:
        """
        Filter elements using a boolean selector iterable.

        Args:
            selectors: Boolean values indicating which elements to keep.
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_("ABCDEF").compress(1, 0, 1, 0, 1, 1).into(list)
        ['A', 'C', 'E', 'F']

        ```
        """
        return self.apply(itertools.compress, selectors)

    def unique(self, key: Callable[[T], Any] | None = None) -> Iter[T]:
        """
        Return only unique elements of the iterable.

        Args:
            key: Function to transform items before comparison. Defaults to None.
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 3]).unique().into(list)
        [1, 2, 3]
        >>> pc.Iter.from_([1, 2, 1, 3]).unique().into(list)
        [1, 2, 3]

        ```
        Uniqueness can be defined by key keyword
        ```python
        >>> pc.Iter.from_(["cat", "mouse", "dog", "hen"]).unique(key=len).into(list)
        ['cat', 'mouse']

        ```
        """
        return self.apply(cz.itertoolz.unique, key=key)

    def take(self, n: int) -> Iter[T]:
        """
        Creates an iterator that yields the first n elements, or fewer if the underlying iterator ends sooner.

        `Iter.take(n)` yields elements until n elements are yielded or the end of the iterator is reached (whichever happens first).

        The returned iterator is either:

        - A prefix of length n if the original iterator contains at least n elements
        - All of the (fewer than n) elements of the original iterator if it contains fewer than n elements.

        Args:
            n: Number of elements to take.
        Example:
        ```python
        >>> import pyochain as pc
        >>> data = [1, 2, 3]
        >>> pc.Iter.from_(data).take(2).into(list)
        [1, 2]
        >>> pc.Iter.from_(data).take(5).into(list)
        [1, 2, 3]

        ```
        """

        return self.apply(partial(cz.itertoolz.take, n))

    def skip(self, n: int) -> Iter[T]:
        """
        Drop first n elements.

        Args:
            n: Number of elements to skip.
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 3]).skip(1).into(list)
        [2, 3]

        ```
        """
        return self.apply(partial(cz.itertoolz.drop, n))

    def unique_justseen(self, key: Callable[[T], Any] | None = None) -> Iter[T]:
        """
        Yields elements in order, ignoring serial duplicates.

        Args:
            key: Function to transform items before comparison. Defaults to None.
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_("AAAABBBCCDAABBB").unique_justseen().into(list)
        ['A', 'B', 'C', 'D', 'A', 'B']
        >>> pc.Iter.from_("ABBCcAD").unique_justseen(str.lower).into(list)
        ['A', 'B', 'C', 'A', 'D']

        ```
        """
        return self.apply(mit.unique_justseen, key=key)

    def unique_in_window(
        self, n: int, key: Callable[[T], Any] | None = None
    ) -> Iter[T]:
        """
        Yield the items from iterable that haven't been seen recently.

        The items in iterable must be hashable.
        Args:
            n: Size of the lookback window.
            key: Function to transform items before comparison. Defaults to None.
        Example:
        ```python
        >>> import pyochain as pc
        >>> iterable = [0, 1, 0, 2, 3, 0]
        >>> n = 3
        >>> pc.Iter.from_(iterable).unique_in_window(n).into(list)
        [0, 1, 2, 3, 0]

        ```
        The key function, if provided, will be used to determine uniqueness:
        ```python
        >>> pc.Iter.from_("abAcda").unique_in_window(3, key=str.lower).into(list)
        ['a', 'b', 'c', 'd', 'a']

        ```
        """
        return self.apply(mit.unique_in_window, n, key=key)

    def extract(self, indices: Iterable[int]) -> Iter[T]:
        """
        Yield values at the specified indices.

        - The iterable is consumed lazily and can be infinite.
        - The indices are consumed immediately and must be finite.
        - Raises IndexError if an index lies beyond the iterable.
        - Raises ValueError for negative indices.

        Args:
            indices: Iterable of indices to extract values from.
        Example:
        ```python
        >>> import pyochain as pc
        >>> text = "abcdefghijklmnopqrstuvwxyz"
        >>> pc.Iter.from_(text).extract([7, 4, 11, 11, 14]).into(list)
        ['h', 'e', 'l', 'l', 'o']

        ```
        """
        return self.apply(mit.extract, indices)

    def every(self, index: int) -> Iter[T]:
        """
        Return every nth item starting from first.

        Args:
            index: Step size for selecting items.
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([10, 20, 30, 40]).every(2).into(list)
        [10, 30]

        ```
        """
        return self.apply(partial(cz.itertoolz.take_nth, index))

    def slice(self, start: int | None = None, stop: int | None = None) -> Iter[T]:
        """
        Return a slice of the iterable.

        Args:
            start: Starting index of the slice. Defaults to None.
            stop: Ending index of the slice. Defaults to None.
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, 2, 3, 4, 5]).slice(1, 4).into(list)
        [2, 3, 4]

        ```
        """

        def _slice(data: Iterable[T]) -> Iterator[T]:
            return itertools.islice(data, start, stop)

        return self.apply(_slice)

    def filter_subclass[U: type[Any], R](
        self: IterWrapper[U], parent: type[R], keep_parent: bool = True
    ) -> Iter[type[R]]:
        """
        Return elements that are subclasses of the given class, optionally excluding the parent class itself.

        Args:
            parent: Parent class to check against.
            keep_parent: Whether to include the parent class itself. Defaults to True.
        Example:
        ```python
        >>> import pyochain as pc
        >>> class A:
        ...     pass
        >>> class B(A):
        ...     pass
        >>> class C:
        ...     pass
        >>> def name(cls: type[Any]) -> str:
        ...     return cls.__name__
        >>>
        >>> data = pc.Seq([A, B, C])
        >>> data.iter().filter_subclass(A).map(name).into(list)
        ['A', 'B']
        >>> data.iter().filter_subclass(A, keep_parent=False).map(name).into(list)
        ['B']

        ```
        """

        def _filter_subclass(
            data: Iterable[type[Any]],
        ) -> Generator[type[R], None, None]:
            if keep_parent:
                return (x for x in data if issubclass(x, parent))
            else:
                return (x for x in data if issubclass(x, parent) and x is not parent)

        return self.apply(_filter_subclass)

    def filter_type[R](self, typ: type[R]) -> Iter[R]:
        """
        Return elements that are instances of the given type.

        Args:
            typ: Type to check against.
        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([1, "two", 3.0, "four", 5]).filter_type(int).into(list)
        [1, 5]

        ```
        """

        def _filter_type(data: Iterable[T]) -> Generator[R, None, None]:
            return (x for x in data if isinstance(x, typ))

        return self.apply(_filter_type)

    def filter_callable(self) -> Iter[Callable[..., Any]]:
        """
        Return only elements that are callable.

        Example:
        ```python
        >>> import pyochain as pc
        >>> pc.Iter.from_([len, 42, str, None, list]).filter_callable().into(list)
        [<built-in function len>, <class 'str'>, <class 'list'>]

        ```
        """

        def _filter_callable(
            data: Iterable[T],
        ) -> Generator[Callable[..., Any], None, None]:
            return (x for x in data if callable(x))

        return self.apply(_filter_callable)

    def filter_map[R](self, func: Callable[[T], R]) -> Iter[R]:
        """
        Apply func to every element of iterable, yielding only those which are not None.

        Args:
            func: Function to apply to each item.
        Example:
        ```python
        >>> import pyochain as pc
        >>> def to_int(s: str) -> int | None:
        ...     return int(s) if s.isnumeric() else None
        >>> elems = ["1", "a", "2", "b", "3"]
        >>> pc.Iter.from_(elems).filter_map(to_int).into(list)
        [1, 2, 3]

        ```
        """
        return self.apply(partial(mit.filter_map, func))
